
{*******************************************************}
{                                                       }
{       Turbo Pascal Version 7.0                        }
{       Multi-Sprites Editor Unit                       }
{                                                       }
{       Copyright (c) 1994,95 by Solar Designer         }
{                                                       }
{*******************************************************}

unit SEditor;
interface
{$R-,B-,V-}
uses
   Objects, Memory,
   Tables, Smooth, Edge, Gray, GraphFns, Scale,
   GViews, GMenus, GRect, GPalette, GMsgBox, GApp,
   GraphDrv, GDrivers, Events, KeyMouse, Utils, General,
   ReadFmt, Hints;

const
   SpriteExt =               'SPR';
   SpriteWildCard =          '*.'+SpriteExt;
   PaletteExt =              'PAL';
   PaletteWildCard =         '*.'+PaletteExt;

   FileBuffer =    1024;
   MaxImageSizeX = 1024;
   MaxImageSizeY = 768;
   MaxImageSize =  64000+4;

   DefaultName =   'UNTITLED.SPR';
   ClipboardName = 'CLIPBRD.SPR';
   DefaultColor =  255;

   MinScale =      1;
   MaxScale =      64;

   MaxUndo =       64;

   PaletteGrid     :TGPoint= (X:4; Y:4);
   PaletteCount    :TGPoint= (X:8; Y:32);

   cpMouseLeft =   #$F0;
   cpMouseRight =  #$F1;

   cpDark =        #$F4;
   cpLight =       #$F5;

   CPaletteBox =
   cpMain+cpLightGray + cpFrame+cpWhite +
   cpMouseLeft+cpWhite + cpMouseRight+cpBlack +
   cpDark+cpBlack + cpLight+cpWhite;

   CInfoBar =
   cpMain+cpLightGray + cpFrame+cpWhite +
   cpText+cpBlack +
   cpDark+cpBlack + cpLight+cpWhite;

   cmRemoveWindow= 4000;

   cmScaleDown =   $F0;
   cmScaleUp =     $F1;

   cmSetColor =    4100;
   cmGetColor =    4101;
   cmFillArea =    4102;

   cmDraggable =   4110;

   cmReplace =     4110;
   cmRectangle =   4111;
   cmBar =         4112;
   cmLine =        4113;

   cmEffectB =     $F4;

   cmFlipX =       $F4;
   cmFlipY =       $F5;

   cmSmooth =      $F6;
   cmEdge =        $F7;
   cmGray =        $F8;

   cmEffectE =     $FA;
   cmScale =       $FB;

   cmInfo =        5000;

   cmClipboard =   $FC;

   cmImageB =      $D8;
   cmNextImage =   $D8;
   cmPrevImage =   $D9;
   cmInsImage =    $DA;
   cmDelImage =    $DB;
   cmImageE =      $DF;

   cmDrawDragging= 6000;
   cmNoDragging =  6001;

   ImageHeaderSize=4+4 + 4 + 1+4 + 4+4 + 4+2 + 2;
   ImageCommonSize=                4+4 + 4+2 + 2;

const
   EditMode        :Word =   cmSetColor;
   WinCount        :Integer= 0;

type
   PWord =         ^Word;

   PUndoItem =     ^TUndoItem;
   TUndoItem =
   object
      Next         :PUndoItem;
      Image        :Pointer;
      Modified     :Boolean;

      constructor Init(ANext           :PUndoItem);
      destructor Done(var List         :PUndoItem);
   end;

   PGEditorField = ^TGEditorField;

   PGEditorImage = ^TGEditorImage;
   TGEditorImage = object

      Next, Prev   :PGEditorImage;
      Owner        :PGEditorField;

      Modified     :Boolean;
      UndoList     :PUndoItem;

      FormatId     :Integer;
      HeaderSize   :Word;
      Header       :Pointer;
      FileSize     :LongInt;
      Size         :TGPoint;

      Data         :Array [0..320*200-1] of Byte;

      procedure Clear;
      procedure Free;

      function  ValidPos(Pos           :TGPoint) :Boolean;
      function  GetColor(Pos           :TGPoint) :Integer;
      procedure SetColor(Pos           :TGPoint;
                         Color         :Integer);
      function  Effect(Cmd             :Word)    :Boolean;
      procedure Scale(Dest             :PGEditorImage);
      procedure ShowUsed;

      procedure StoreUndo;
      procedure Undo;
   end;

   TGEditorField =
   object(TView)

      Image        :PGEditorImage;

      ListModified :Boolean;

      Grid         :TGPoint;
      ShowSteps    :Boolean;

      constructor Init(var R           :TGRect;
                       AnImage         :PGEditorImage);

      procedure ArrangeGrid(var R      :TGRect);
      procedure CalcBounds(var R       :TGRect;
                           Delta       :TGPoint); virtual;
      procedure Draw; virtual;
      procedure DrawPos(Pos            :TGPoint);
      procedure HandleEvent(var Event  :TEvent); virtual;

      function  Modified                         :Boolean;

   private
      StartPos     :TGPoint;
      LastPos      :TGPoint;
      StartButtons :Integer;
   end;

   PGEditor =      ^TGEditor;
   TGEditor =
   object(TWindow)

      Field        :PGEditorField;
      Image,
      FirstImage   :PGEditorImage;
      HScrollBar,
      VScrollBar   :PScrollBar;

      IsClipboard  :Boolean;

      constructor Init(var Bounds      :TGRect;
                       const FileName  :String;
                       ANumber         :Integer);
      constructor InitNew(var Bounds   :TGRect;
                          ANumber      :Integer;
                          ImageSize    :TGPoint;
                          FormatId     :Integer;
                          AHeader      :Pointer);
      destructor Done; virtual;

      function  GetMenu(NextItem       :Pointer) :Pointer; virtual;

      procedure UpdateScrollBars;
      procedure UpdateCommands;
      procedure UpdateInfo;

      procedure HandleEvent(var Event  :TEvent); virtual;
      procedure SetState(AState        :Word;
                         Enable        :Boolean); virtual;

      function  GetTitle(MaxSize       :Integer) :TTitleStr; virtual;
      function  Valid(Command          :Word)    :Boolean; virtual;

      procedure FreeImage;
      procedure FreeImages;
      procedure NewImage(ImageSize     :TGPoint;
                         FormatId      :Integer;
                         AHeader       :Pointer);
      function  ConvertImage(FormatId  :Integer;
                             AHeader   :Pointer) :Boolean;
      procedure LoadImage;
      function  SaveImage(FileName     :PString) :Boolean;
   end;

   TColorView =
   object(TView)
      procedure DrawColor(var Bounds   :TGRect;
                          Color        :Integer);
   end;

   PPaletteBox =   ^TPaletteBox;
   TPaletteBox =
   object(TColorView)

      CLeft,
      CRight,
      CCurrent     :Integer;
      Used         :Set of Byte;
      UsageShown   :Boolean;

      constructor Init(var Right       :TGRect);

      function  GetPalette                       :TPalette; virtual;

      procedure Draw; virtual;
      procedure HandleEvent(var Event  :TEvent); virtual;

      procedure DrawCurrent(Color      :Integer);
      procedure Update(Mouse           :TGPoint);
   end;

   PInfoBar =      ^TInfoBar;
   TInfoBar =
   object(TColorView)

      Modified     :Boolean;
      Pos          :TGPoint;
      ImageNum,
      ImageCnt     :Integer;
      Color        :Integer;
      CorrectData  :Boolean;

      function  GetPalette                       :TPalette; virtual;

      procedure Draw; virtual;
      procedure DrawPos;
   end;

var
   PaletteBox      :PPaletteBox;
   InfoBar         :PInfoBar;
   PaletteRGB      :TRGBPalette;
   Clipboard       :PGEditor;

implementation
uses
   LPalette;

{ TUndoItem }

constructor TUndoItem.Init;
begin
   Next:=ANext;
end;

destructor TUndoItem.Done;
begin
   List:=Next;
   if Image<>nil then DisposeCache(Image);
end;

{ TGEditorImage }

procedure TGEditorImage.Clear;
begin
   FillChar(Self, ImageHeaderSize, 0);
   Prev:=@Self; Next:=@Self;
end;

procedure TGEditorImage.Free;
begin
   while UndoList<>nil do Dispose(UndoList, Done(UndoList));
   FreeMem(Header, HeaderSize);
   FreeMem(@Self, FileSize + (ImageHeaderSize-4));
end;

function  TGEditorImage.ValidPos;
begin
   ValidPos:=(Pos.X>=0) and (Pos.Y>=0) and (Pos.X<Size.X) and (Pos.Y<Size.Y);
end;

function  TGEditorImage.GetColor;
begin
   if ValidPos(Pos)
   then GetColor:=Data[Word(Pos.Y)*Size.X+Word(Pos.X)] else GetColor:=0;
end;

procedure TGEditorImage.SetColor;
begin
   if ValidPos(Pos) then Data[Word(Pos.Y)*Size.X+Word(Pos.X)]:=Color;
end;

function  TGEditorImage.Effect;
var
   Buf, Src        :Pointer;
begin
   Buf:=MemAlloc(FileSize);
   if Buf=nil then Application^.OutOfMemory else
   begin
      case Cmd of
         cmFlipX:  GraphDrv.TurnBlockX(@Size, Buf);
         cmFlipY:  GraphDrv.TurnBlockY(@Size, Buf);

         cmSmooth..cmGray:
         begin
            FreeMem(Buf, FileSize);

            if Test8086<2 then
            begin
               MessageBox(^M^C'At least an 80386'^M^M^C'CPU is required.', nil, mfError+mfOkButton);
               Effect:=False; Exit;
            end;

            Src:=MemAllocSeg(320*200); Buf:=MemAllocSeg(320*200);

            Effect:=(Src<>nil) and (Buf<>nil);

            if (Src=nil) or (Buf=nil) then Application^.OutOfMemory else
            begin
               CalcTables(Cmd<>cmSmooth);

               Application^.HelpCtx:=hcWait;
               StatusLine^.Update;

               if (Cmd<>cmSmooth) or (cRGB2C<>nil) then
               begin
                  PutImageBuf(0, 0, @Size, PtrRec(Src).Seg);
                  PutImageBuf(0, 0, @Size, PtrRec(Buf).Seg);

                  case Cmd of
                     cmSmooth: SmoothArea(1, 1, Size.X-2, Size.Y-2, PtrRec(Src).Seg, PtrRec(Buf).Seg);
                     cmEdge:   EdgeArea(1, 1, Size.X-2, Size.Y-2, PtrRec(Src).Seg, PtrRec(Buf).Seg);
                     cmGray:   GrayArea(0, 0, Size.X, Size.Y, PtrRec(Src).Seg, PtrRec(Buf).Seg);
                  end;

                  GetImageBuf(0, 0, Size.X, Size.Y, @Size, PtrRec(Buf).Seg);
               end else Application^.OutOfMemory;

               Application^.HelpCtx:=hcNoContext;
            end;

            if Buf<>nil then FreeMem(Buf, 320*200);
            if Src<>nil then FreeMem(Src, 320*200);

            Exit;
         end;
      end;

      Move(Buf^, Size, FileSize);
      FreeMem(Buf, FileSize);
   end;
   Effect:=(Buf<>nil);
end;

procedure TGEditorImage.Scale;
begin
   ScaleImage(@Size, @Dest^.Size, Dest^.Size);
end;

procedure TGEditorImage.ShowUsed;
begin
   asm
      les  di,Self
      mov  ax,es:[di].TGEditorImage.Size.X
      mul  es:[di].TGEditorImage.Size.Y
      xchg ax,cx
      lea  si,[di].TGEditorImage.Data
      les  di,PaletteBox
      push ds
      push word ptr Self+2
      pop  ds
      cld
@@Loop:
      xor  ax,ax
      lodsb

      mov  bx,ax
      shr  bx,3
      push cx
      and  al,7
      xchg cx,ax
      mov  al,1
      shl  ax,cl
      pop  cx

      or   byte ptr es:[di].TPaletteBox.Used[bx],al
      loop @@Loop
      pop  ds
   end;

   PaletteBox^.UsageShown:=True;
   PaletteBox^.DrawView;
end;

procedure TGEditorImage.StoreUndo;
var
   Count           :Integer;
   Current, U1, U2 :PUndoItem;
begin
   Count:=0; Current:=UndoList;
   while Current<>nil do
   begin
      if Current^.Image<>nil then Inc(Count);
      U2:=U1; U1:=Current; Current:=Current^.Next;
   end;

   if Count>=MaxUndo then
   begin
      Dispose(U1, Done(U1)); U2^.Next:=nil;
   end;

   New(UndoList, Init(UndoList));
   NewCache(UndoList^.Image, Size.X*Size.Y);

   if UndoList^.Image<>nil then
   begin
      Move(Data, UndoList^.Image^, Size.X*Size.Y);
      UndoList^.Modified:=Modified;
   end else Dispose(UndoList, Done(UndoList));
end;

procedure TGEditorImage.Undo;
begin
   while (UndoList<>nil) and (UndoList^.Image=nil) do Dispose(UndoList, Done(UndoList));
   if UndoList<>nil then
   begin
      Move(UndoList^.Image^, Data, Size.X*Size.Y);
      Modified:=UndoList^.Modified;
      Dispose(UndoList, Done(UndoList));
      Owner^.DrawView;

      InfoBar^.Modified:=Owner^.Modified;
      if not InfoBar^.Modified then InfoBar^.DrawPos;
   end
   else MessageBox(^M^M^C'Unable to undo.', nil, mfError+mfOKButton);
end;

{ TGEditorField }

constructor TGEditorField.Init;
begin
   Inherited Init(R);

   Image:=AnImage; Image^.Owner:=@Self; ArrangeGrid(R);
   GrowMode:=gfGrowHiX+gfGrowHiY;
   EventMask:=EventMask or (evMouseMove+evMouseDown+evMouseUp + evBroadcast);

   StartPos.X:=MaxInt; StartPos.Y:=MaxInt;
   LastPos.X:=MaxInt; LastPos.Y:=MaxInt;
end;

procedure TGEditorField.CalcBounds;
begin
   Inherited CalcBounds(R, Delta);
   PGEditor(Owner)^.UpdateScrollBars;
   PaletteBox^.InitClipping;
end;

procedure TGEditorField.ArrangeGrid;
var
   S               :TGPoint;
begin
   if Image=nil then LongInt(Grid):=$00010001 else
   begin
      R.GetSize(S);

      Grid.X:=Max(S.X div Image^.Size.X, 4);
      Grid.Y:=Max(S.Y div Image^.Size.Y, 4);

      if (Image^.Size.X>64) or (Image^.Size.Y>64) then LongInt(Grid):=$00010001;

      Grid.X:=Min(Grid.X, Grid.Y); Grid.Y:=Grid.X;
      if LongInt(Grid)<>$00010001 then
      if ScreenWidth<=ScreenHeight then Grid.Y:=Grid.Y shl 1 else
      if ScreenHeight<200 then Grid.X:=Grid.X shl 1;
   end;

   if Owner<>nil then PGEditor(Owner)^.UpdateScrollBars;
end;

procedure TGEditorField.Draw;
var
   R               :TGRect;
   Pos, ScrollPos  :TGPoint;
begin
   Message(@Self, evBroadcast, cmDrawDragging, nil);

   ScrollPos.X:=PGEditor(Owner)^.HScrollBar^.Value;
   ScrollPos.Y:=PGEditor(Owner)^.VScrollBar^.Value;

   Pos.X:=-ScrollPos.X * Grid.X; Pos.Y:=-ScrollPos.Y * Grid.Y;
   DrawImage(Pos, Grid, @(Image^.Size), NullColor);
   Pos.Move(Image^.Size.X * Grid.X, Image^.Size.Y * Grid.Y);

   GetExtent(R); R.A.Y:=Pos.Y;
   if R.A.Y<=R.B.Y then DrawBar(R, GetColor(cpMain));
   R.A.X:=Pos.X; R.A.Y:=0; R.B.Y:=Pos.Y-1;
   if R.A.X<=R.B.X then DrawBar(R, GetColor(cpMain));

   Message(@Self, evBroadcast, cmDrawDragging, nil);
end;

procedure TGEditorField.DrawPos;
var
   R, Clip         :TGRect;
begin
   if Pos.Rect(0, 0, Image^.Size.X-1, Image^.Size.Y-1) then
   begin
      R.A.X:=(Pos.X-PGEditor(Owner)^.HScrollBar^.Value)*Grid.X;
      R.A.Y:=(Pos.Y-PGEditor(Owner)^.VScrollBar^.Value)*Grid.Y;
      R.SetSize(Grid);
      DrawBar(R, Image^.GetColor(Pos));
   end;
end;

var
   CommonField     :PGEditorField;

procedure DoPutPixel(x, y, Color       :Integer); far;
var
   Pos             :TGPoint;
begin
   Pos.X:=x; Pos.Y:=y;
   with CommonField^ do
   begin
      Image^.SetColor(Pos, Color);
      if ShowSteps then DrawPos(Pos);
   end;
end;

procedure TGEditorField.HandleEvent;
var
   Pos             :TGPoint;
   CurrentColor    :Integer;
   R               :TGRect;

procedure FloodFill(cf                 :Integer);
type
   TCoordArray =   Array [1..$4000] of Integer;
var
   SX, SY, SXNew, SYNew,
   Temp, MarkMem                       :^TCoordArray;
   x, y, cb,
   i, n, NNew, MaxX, MaxY, MaxIndex    :Integer;
   ArrayMem                            :Word;
   BufSize                             :LongInt;
   P                                   :TGPoint;

procedure FillPixel(x, y               :Integer);
begin
   if NNew>=MaxIndex then Exit;
   P.X:=x; P.Y:=y;
   Image^.SetColor(P, cf); DrawPos(P);
   Inc(NNew); SXNew^[NNew]:=x; SYNew^[NNew]:=y;
end;

function  GetPixel(x, y                :Integer) :Integer;
begin
   P.X:=x; P.Y:=y;
   GetPixel:=Image^.GetColor(P);
end;

begin
   x:=Pos.X; y:=Pos.Y;

   BufSize:=MaxAvail shr 1;
   if BufSize>64000 then BufSize:=64000;

   if BufSize<4096 then
   begin
      Application^.OutOfMemory; Exit;
   end;

   MaxIndex:=BufSize shr 3; ArrayMem:=MaxIndex shl 1;

   Mark(MarkMem);
   GetMem(SX, ArrayMem); GetMem(SY, ArrayMem);
   GetMem(SXNew, ArrayMem); GetMem(SYNew, ArrayMem);

   GetExtent(R);
   HideMouseLocal(R);

   Application^.HelpCtx:=hcWait;
   StatusLine^.Update;

   MaxX:=Image^.Size.X-1; MaxY:=Image^.Size.Y-1;
   NNew:=0;
   cb:=GetPixel(x, y);
   if cb<>cf then
   begin
      FillPixel(x, y);
      repeat
         n:=NNew; NNew:=0;
         Temp:=SX; SX:=SXNew; SXNew:=Temp; Temp:=SY; SY:=SYNew; SYNew:=Temp;
         for i:=1 to n do
         begin
            x:=sx^[i]; y:=sy^[i];
            if (x>0) and (GetPixel(x-1, y)=cb) then FillPixel(x-1, y);
            if (x<MaxX) and (GetPixel(x+1, y)=cb) then FillPixel(x+1, y);
            if (y>0) and (GetPixel(x, y-1)=cb) then FillPixel(x, y-1);
            if (y<MaxY) and (GetPixel(x, y+1)=cb) then FillPixel(x, y+1);
         end;
      until NNew=0;
   end;

   Application^.HelpCtx:=hcNoContext;

   ShowMouseRect;
   Release(MarkMem);

   Image^.Modified:=True;
end;

procedure DrawDragObject(Pos           :TGPoint);
var
   R               :TGRect;
   S               :TGPoint;
begin
   S.X:=PGEditor(Owner)^.HScrollBar^.Value;
   S.Y:=PGEditor(Owner)^.VScrollBar^.Value;

   R.A.X:=(StartPos.X-S.X)*Grid.X; R.A.Y:=(StartPos.Y-S.Y)*Grid.Y;
   R.B.X:=(Pos.X-S.X)*Grid.X; R.B.Y:=(Pos.Y-S.Y)*Grid.Y;
   R.Move(Grid.X shr 1, Grid.Y shr 1);

   GraphDrv.DriverPtr^.PutMode:=GraphDrv.XorPut;
   case EditMode of
      cmReplace, cmRectangle, cmBar, cmCopy:
         if (R.A.X=R.B.X) or (R.A.Y=R.B.Y)
         then DrawLine(R, DragColor) else DrawRect(R, DragColor);
      cmPaste:
      begin
         R.A.X:=R.B.X+(Clipboard^.Image^.Size.X-1)*Grid.X;
         R.A.Y:=R.B.Y+(Clipboard^.Image^.Size.Y-1)*Grid.Y;
         DrawRect(R, DragColor);
      end;
      cmLine:
         DrawLine(R, DragColor);
   end;
   GraphDrv.DriverPtr^.PutMode:=GraphDrv.NormalPut;
end;

procedure DoColor(var Color            :Integer);

procedure DoLine(x1, y1, x2, y2        :Integer);
begin
   CommonField:=@Self;
   GViews.DrawAnyLine(x1, y1, x2, y2, Color, @DoPutPixel, False, $FFFF);
end;

procedure DoReplace;
var
   SrcColor,
   DstColor, Y     :Integer;
   R, S            :TGRect;
   Sz              :TGPoint;
   DataPtr         :PChar;
begin
   DstColor:=Color;
   if DstColor=PaletteBox^.CLeft
   then SrcColor:=PaletteBox^.CRight
   else SrcColor:=PaletteBox^.CLeft;
   if DstColor=SrcColor then Exit;

   S.A:=StartPos; S.B:=Pos; S.Arrange;
   LongInt(R.A):=0; R.SetSize(Image^.Size); Inc(R.B.X); Inc(R.B.Y);
   S.Intersect(R);
   S.GetSize(Sz);

   DataPtr:=PChar(@Image^.Data)+Image^.Size.X*S.A.Y+S.A.X;
   for Y:=1 to Sz.Y do
   begin
      asm
         les  di,DataPtr
         cld
         mov  cx,Sz.X
         mov  al,byte ptr SrcColor
         mov  dl,byte ptr DstColor
@@Next:
         repne scasb
         jcxz @@Done
         mov  es:[di-1],dl
         jmp  @@Next
@@Done:
      end;
      Inc(DataPtr, Image^.Size.X);
   end;

   Owner^.Clip.A.X:=Origin.X+(S.A.X-PGEditor(Owner)^.HScrollBar^.Value)*Grid.X;
   Owner^.Clip.A.Y:=Origin.Y+(S.A.Y-PGEditor(Owner)^.VScrollBar^.Value)*Grid.Y;
   Owner^.Clip.B.X:=Owner^.Clip.A.X+Sz.X*Grid.X;
   Owner^.Clip.B.Y:=Owner^.Clip.A.Y+Sz.Y*Grid.Y;
   DrawView;
   Owner^.GetExtent(Owner^.Clip);

   Image^.Modified:=True;
end;

procedure DoCopy;
var
   R, S            :TGRect;
   Sz              :TGPoint;
   Source, Dest    :PChar;
   Old, Save       :PGEditor;
   OldState        :Word;
begin
   Old:=Clipboard;

   if Old=nil
   then Desktop^.Background^.GetExtent(R)
   else Old^.GetBounds(R);

   S.A:=StartPos; S.B:=Pos; S.Arrange; S.GetSize(Sz);
   New(Clipboard, InitNew(R, wnNoNumber, Sz, 1, nil));
   if Clipboard=nil then
   begin
      if Old<>nil then Old^.Free; Exit;
   end;

   Source:=PChar(@Image^.Data)+Image^.Size.X*S.A.Y+S.A.X;
   Dest:=@Clipboard^.Image^.Data;
   while Sz.Y>0 do
   begin
      Move(Source^, Dest^, Sz.X);
      Inc(Source, Image^.Size.X);
      Inc(Dest, Sz.X);
      Dec(Sz.Y);
   end;

   DisposeStr(Clipboard^.Title); Clipboard^.Title:=NewStr(ClipboardName);
   Clipboard^.IsClipboard:=True;
   Clipboard^.Hide;

   if Old<>nil then
   begin
      OldState:=Old^.State;
      Inc(WinCount);
      Save:=Clipboard; Old^.Free; Clipboard:=Save;
   end else OldState:=0;

   Desktop^.Insert(Clipboard);
   if OldState and sfVisible<>0 then
   begin
      Clipboard^.Show;
      if OldState and sfFocused=0 then Desktop^.SelectNext(True);
   end;

   EnableCommands([cmPaste, cmClipboard]);
end;

procedure DoPaste;
var
   R, S            :TGRect;
   Sz              :TGPoint;
   y               :Integer;
   Source, Dest    :PChar;

procedure MoveColor(Source, Dest       :Pointer;
                    Count, Color       :Word); assembler;
asm
   push ds
   lds  si,Source
   les  di,Dest
   mov  cx,Count
   mov  dl,byte ptr Color
   cld
@@Loop:
   lodsb
   cmp  al,dl
   je   @@Skip
   stosb
   loop @@Loop
   jmp  @@Exit
@@Skip:
   inc  di
   loop @@Loop
@@Exit:
   pop  ds
end;

begin
   if PGEditor(Owner)^.IsClipboard then
   begin
      MessageBox(^M^C'Unable to paste'^M^M^C'to the Clipboard.',
         nil, mfError+mfOkButton);
      Exit;
   end;

   S.A:=Pos; S.SetSize(Clipboard^.Image^.Size);
   LongInt(R.A):=0; R.SetSize(Image^.Size);
   S.Intersect(R);
   S.GetSize(Sz);

   Source:=@Clipboard^.Image^.Data;
   Dest:=PChar(@Image^.Data)+Image^.Size.X*Pos.Y+Pos.X;
   for y:=1 to Sz.Y do
   begin
      if Event.Buttons=mbRightButton
      then MoveColor(Source, Dest, Sz.X, Color)
      else Move(Source^, Dest^, Sz.X);
      Inc(Source, Clipboard^.Image^.Size.X);
      Inc(Dest, Image^.Size.X);
   end;

   Owner^.Clip.A.X:=Origin.X+(Pos.X-PGEditor(Owner)^.HScrollBar^.Value)*Grid.X;
   Owner^.Clip.A.Y:=Origin.Y+(Pos.Y-PGEditor(Owner)^.VScrollBar^.Value)*Grid.Y;
   Owner^.Clip.B.X:=Owner^.Clip.A.X+Sz.X*Grid.X;
   Owner^.Clip.B.Y:=Owner^.Clip.A.Y+Sz.Y*Grid.Y;
   DrawView;
   Owner^.GetExtent(Owner^.Clip);

   Image^.Modified:=True;
end;

begin
   case EditMode of
      cmSetColor:
      begin
         Image^.SetColor(Pos, Color);
         if Color<>CurrentColor then
         begin
            Image^.Modified:=True; DrawPos(Pos); CurrentColor:=Color;
         end;
      end;
      cmGetColor:
      begin
         Color:=Image^.GetColor(Pos);
         EditMode:=cmSetColor;
         InfoBar^.DrawView;
      end;
      cmFillArea:
      begin
         repeat until not MouseEvent(Event, evMouseUp);
         FloodFill(Color);
      end;
      cmPaste:
      begin
         if LastPos.X<>MaxInt then DrawDragObject(LastPos);
         LastPos.X:=MaxInt;
         DoPaste;
      end;
      else
      begin
         if LastPos.X<>MaxInt then DrawDragObject(LastPos);
         LastPos.X:=MaxInt;

         Image^.Modified:=True;

         if StartPos.X<>MaxInt then
         case EditMode of
            cmReplace:
               DoReplace;
            cmRectangle:
            begin
               ShowSteps:=True;
               DoLine(StartPos.X, StartPos.Y, Pos.X, StartPos.Y);
               DoLine(Pos.X, StartPos.Y, Pos.X, Pos.Y);
               DoLine(Pos.X, Pos.Y, StartPos.X, Pos.Y);
               DoLine(StartPos.X, Pos.Y, StartPos.X, StartPos.Y);
            end;
            cmBar:
            begin
               Application^.HelpCtx:=hcWait;
               StatusLine^.Update;

               ShowSteps:=False;
               ArrangeInts(StartPos.Y, Pos.Y);
               while StartPos.Y<=Pos.Y do
               begin
                  DoLine(StartPos.X, StartPos.Y, Pos.X, StartPos.Y);
                  Inc(StartPos.Y);
               end;
               DrawView;

               Application^.HelpCtx:=hcNoContext;
            end;
            cmLine:
            begin
               ShowSteps:=True;
               DoLine(StartPos.X, StartPos.Y, Pos.X, Pos.Y);
            end;
            cmCopy:
            begin
               Image^.Modified:=False;

               DoCopy;
               if Clipboard<>nil then
               begin
                  EditMode:=cmPaste;
                  InfoBar^.DrawView;
               end;
            end;
         end;
         StartPos.X:=MaxInt;
      end;
   end;
end;

procedure DrawDragging;
begin
   if Exposed then
   if LastPos.X<>MaxInt then DrawDragObject(LastPos);
end;

begin
   Inherited HandleEvent(Event);

   if Event.What=evBroadcast then
   case Event.Command of
      cmDrawDragging:
         DrawDragging;
      cmReleasedFocus, cmNoDragging:
      begin
         DrawDragging;
         LastPos.X:=MaxInt; StartPos.X:=MaxInt;
      end;
   end;

   if (Event.What and (evMouseDown+evMouseUp+evMouseMove)<>0) and
      (Owner^.State and sfFocused<>0) then
   begin
      if (EditMode<>cmCopy) and
         (((EditMode<cmDraggable) and (Event.What=evMouseDown)) or
         ((EditMode>=cmDraggable) and (Event.What=evMouseUp))) then Image^.StoreUndo;

      MakeLocal(Event.Where, Event.Where);
      Pos.X:=Event.Where.X div Grid.X; Pos.Y:=Event.Where.Y div Grid.Y;

      Inc(Pos.X, PGEditor(Owner)^.HScrollBar^.Value);
      Inc(Pos.Y, PGEditor(Owner)^.VScrollBar^.Value);

      if (Event.What=evMouseDown) and (not Image^.ValidPos(Pos)) then Exit;

      CurrentColor:=Image^.GetColor(Pos);

      if (((EditMode<cmDraggable) and (EditMode<>cmCopy)) and (Event.What<>evMouseUp)) or
         (((EditMode>=cmDraggable) or (EditMode=cmCopy)) and (Event.What=evMouseUp)) then
      begin
         if Event.What=evMouseUp then Event.Buttons:=StartButtons;
         if Event.Buttons<>0 then
         if Event.Buttons and mbRightButton=0
         then DoColor(PaletteBox^.CLeft) else DoColor(PaletteBox^.CRight);
      end;

      if (EditMode>=cmDraggable) or (EditMode=cmCopy) or (EditMode=cmPaste) then
      case Event.What of
         evMouseDown:
         if Image^.ValidPos(Pos) then
         begin
            StartPos:=Pos; LastPos.X:=MaxInt;
            StartButtons:=Event.Buttons;
         end;
         evMouseMove:
         if (EditMode=cmPaste) or
            ((StartPos.X<>MaxInt) and (LongInt(Pos)<>LongInt(LastPos))) then
         begin
            if LastPos.X<>MaxInt then DrawDragObject(LastPos);
            DrawDragObject(Pos);
            LastPos:=Pos;
         end;
      end;

      if Image^.ValidPos(Pos) and
         ((not InfoBar^.CorrectData) or
         (LongInt(InfoBar^.Pos)<>LongInt(Pos)) or
         (InfoBar^.Color<>CurrentColor)) then
      begin
         InfoBar^.Modified:=Modified;
         InfoBar^.Pos:=Pos; InfoBar^.Color:=CurrentColor;
         InfoBar^.CorrectData:=True;
         InfoBar^.DrawPos;
      end;

      ClearEvent(Event);
   end;
end;

function  TGEditorField.Modified;
var
   Start, Current  :PGEditorImage;
begin
   Modified:=ListModified;

   if Image=nil then Exit;
   Start:=Image; Current:=Start;
   repeat
      if Current^.Modified then Modified:=True;
      Current:=Current^.Next;
   until Current=Start;
end;

{ TGEditor }

constructor TGEditor.Init;
var
   R               :TGRect;
   FileSize        :LongInt;
   PaletteFile     :PathStr;
   S               :TDosStream;
   Start, Cur      :PGEditorImage;
begin
   Inherited Init(Bounds, FileName, ANumber);
   if FileName<>'' then
   begin
      LoadImage; if Image=nil then Fail;
   end else Title:=NewStr(DefaultName);

   PaletteFile:=Copy(FileName, 1, Pos('.', FileName))+PaletteExt;
   S.Init(PaletteFile, stOpenRead);
   if S.GetSize=768 then LoadPalette(S, pfRaw);
   S.Done;

   HScrollBar:=StandardScrollBar(sbHorizontal+sbHandleKeyboard);
   VScrollBar:=StandardScrollBar(sbVertical+sbHandleKeyboard);
   HScrollBar^.SetStep(8, 1); VScrollBar^.SetStep(8, 1);

   GetInterior(R);
   New(Field, Init(R, Image));

   if Image<>nil then
   begin
      Start:=Image; Cur:=Start;
      repeat
         Cur^.Owner:=Field;

         Cur:=Cur^.Next;
      until Cur=Start;
   end;

   Insert(Field);
end;

constructor TGEditor.InitNew;
var
   R               :TGRect;
begin
   Init(Bounds, '', ANumber);
   NewImage(ImageSize, FormatId, AHeader);
   if Image=nil then Fail;
   FirstImage:=Image;
   Field^.Image:=Image; Image^.Owner:=Field;
   Field^.GetExtent(R); Field^.ArrangeGrid(R);
end;

destructor TGEditor.Done;
begin
   FreeImages;
   Inherited Done;
   if IsClipboard then
   begin
      Clipboard:=nil;
      DisableCommands([cmPaste, cmClipboard]);
      if EditMode=cmPaste then
      begin
         EditMode:=cmSetColor; InfoBar^.DrawView;
      end;
   end;
   Message(Application, evCommand, cmRemoveWindow, @Self);
end;

function  TGEditor.GetMenu;
begin
   GetMenu:=
     Inherited GetMenu(
     NewLine(
     NewItem('Scale ~d~own', 'Gray -', kbGrayMinus, cmScaleDown, hcScaleDown,
     NewItem('Scale ~u~p', 'Gray +', kbGrayPlus, cmScaleUp, hcScaleUp,
     NewLine(
     NewItem('~N~ext image', 'Ctrl+End', kbCtrlEnd, cmNextImage, hcNextImage,
     NewItem('~P~revious image', 'Ctrl+Home', kbCtrlHome, cmPrevImage, hcPrevImage,
     NewLine(
     NewItem('~I~nformation...', '', kbNoKey, cmInfo, hcInfo,
   nil)))))))));
end;

procedure TGEditor.UpdateScrollBars;
var
   R               :TGRect;
begin
   if Image=nil then
   begin
      HScrollBar^.SetRange(0, 0); VScrollBar^.SetRange(0, 0); Exit;
   end;
   GetInterior(R);
   HScrollBar^.SetRange(0, Image^.Size.X - R.SizeX div Field^.Grid.X);
   VScrollBar^.SetRange(0, Image^.Size.Y - R.SizeY div Field^.Grid.Y);
   HScrollBar^.Show; VScrollBar^.Show;

   UpdateCommands;
end;

procedure TGEditor.UpdateCommands;
var
   Cmds            :TCommandSet;
begin
   if State and sfSelected = 0 then Exit;
   DisableCommands([cmScaleDown, cmScaleUp]);
   Cmds:=[];
   if Field^.Grid.X>MinScale then Cmds:=Cmds+[cmScaleDown];
   if Field^.Grid.X<MaxScale then Cmds:=Cmds+[cmScaleUp];
   EnableCommands(Cmds);
end;

procedure TGEditor.UpdateInfo;
var
   Start, Cur      :PGEditorImage;
   Num, Cnt        :Integer;
begin
   Start:=FirstImage; Cur:=Start; Num:=0; Cnt:=0;
   if Start<>nil then
   repeat
      Inc(Cnt);
      if Cur=Image then Num:=Cnt;
      Cur:=Cur^.Next;
   until Cur=Start;
   InfoBar^.ImageNum:=Num; InfoBar^.ImageCnt:=Cnt;
end;

procedure TGEditor.HandleEvent;

procedure HandleImageCommand;
var
   LastImage,
   SaveImage       :PGEditorImage;
begin
   LastImage:=Image;

   case Event.Command of
      cmNextImage:
         Image:=Image^.Next;
      cmPrevImage:
         Image:=Image^.Prev;
      cmInsImage:
      begin
         SaveImage:=MemAlloc(Image^.FileSize + (ImageHeaderSize-4));
         if SaveImage=nil then
         begin
            Application^.OutOfMemory; Exit;
         end;

         SaveImage^.Clear;
         Move(Image^.FormatId, SaveImage^.FormatId, ImageCommonSize);
         SaveImage^.Header:=MemAlloc(SaveImage^.HeaderSize);
         if SaveImage^.Header=nil then
         begin
            FreeMem(SaveImage, SaveImage^.FileSize + (ImageHeaderSize-4));
            Application^.OutOfMemory; Exit;
         end;
         Move(Image^.Header^, SaveImage^.Header^, SaveImage^.HeaderSize);
         Move(Image^.Data, SaveImage^.Data, SaveImage^.FileSize-4);

         Image^.Owner:=Field;

         SaveImage^.Prev:=Image;
         SaveImage^.Next:=Image^.Next;
         Image^.Next^.Prev:=SaveImage;
         Image^.Next:=SaveImage;
         Image:=SaveImage;

         Field^.ListModified:=True;
      end;
      cmDelImage:
      if Image<>Image^.Next then
      if MessageBox(^M^C'Delete the current image?',
         nil, mfConfirmation + mfOKButton + mfCancelButton) = cmOK then
      begin
         SaveImage:=Image^.Next;
         SaveImage^.Prev:=Image^.Prev;
         Image^.Prev^.Next:=SaveImage;
         FreeImage; Image:=SaveImage;

         Field^.ListModified:=True;
      end;
   end;

   if Image<>LastImage then
   begin
      Field^.Image:=Image;
      UpdateInfo; InfoBar^.DrawView;
      Field^.DrawView;
   end;
end;

begin
   Inherited HandleEvent(Event);
   if Event.What=evBroadcast then
   case Event.Command of
      cmScrollBarChanged:
         if Field<>nil then Field^.DrawView;
      cmReceivedFocus:
         UpdateScrollBars;

      cmImageB..cmImageE:
         HandleImageCommand;
   end;
end;

procedure TGEditor.SetState;
begin
   Inherited SetState(AState, Enable);
   if (AState = sfSelected) or ((AState = sfExposed) and
      (State and sfSelected <> 0)) then
   begin
      UpdateCommands;
      if Enable then UpdateInfo;

      InfoBar^.Modified:=Field^.Modified;
      InfoBar^.DrawPos;
   end;
end;

function  TGEditor.GetTitle;
var
   i               :Integer;
begin
   if IsClipboard then GetTitle:='Clipboard' else
   if Title=nil then GetTitle:='Untitled' else
   begin
      if Length(Title^)<=MaxSize then GetTitle:=Title^ else
      begin
         i:=Length(Title^);
         while (i>0) and (Title^[i]<>'\') do Dec(i);
         if Length(Title^)-i<=MaxSize then GetTitle:=Copy(Title^, i+1, 255)
         else GetTitle:='';
      end;
   end;
end;

function  TGEditor.Valid;
begin
   if (Command=cmClose) or (Command=cmQuit) then
   if Field^.Modified then
   begin
      Command:=MessageBox(
         ^M^C'The image '+GetTitle(12)+
         ^M^M^C'has been modified. Save?',
         nil, mfConfirmation+mfYesButton+mfNoButton+mfCancelButton);
      if Command=cmYes then
      if not SaveImage(Title) then Command:=cmCancel;
      Valid:=(Command<>cmCancel);
   end else Valid:=True
   else Valid:=Inherited Valid(Command);
end;

procedure TGEditor.FreeImage;
begin
   if Image<>nil then
   begin
      if FirstImage=Image then FirstImage:=Image^.Next;
      if FirstImage=Image then FirstImage:=nil;
      Image^.Free;
      Image:=nil;
      if Field<>nil then Field^.Image:=nil;
   end;
end;

procedure TGEditor.FreeImages;
var
   Start, Cur      :PGEditorImage;
begin
   if Image=nil then Exit;
   Start:=Image; Cur:=Start;
   repeat
      Cur:=Cur^.Next;
      FreeImage; Image:=Cur;
   until Cur=Start;
   Image:=nil;
end;

procedure TGEditor.NewImage;
var
   FileSize        :Word;
begin
   FreeImages;
   FileSize:=Word(ImageSize.X)*ImageSize.Y + 4;
   Image:=MemAlloc(FileSize + (ImageHeaderSize-4));
   if Image=nil then
   begin
      Application^.OutOfMemory; Exit;
   end;
   Image^.Clear;
   Image^.FileSize:=FileSize; Image^.Size:=ImageSize;

   if not ConvertImage(FormatId, AHeader) then
   begin
      FreeMem(Image, FileSize + (ImageHeaderSize-4)); Exit;
   end;

   FillChar(Image^.Data, FileSize-4, DefaultColor);
end;

function  TGEditor.ConvertImage;
var
   Header          :PByteArray;
   FileSizeH       :Word;
   ImageSizeH      :TGPoint;
begin
   ConvertImage:=False;
   with Formats[FormatId] do
   begin
      Header:=MemAlloc(HeaderSize);
      if Header=nil then
      begin
         Application^.OutOfMemory; Exit;
      end;
      if Image^.Header<>nil then FreeMem(Image^.Header, Image^.HeaderSize);
      Image^.Header:=Header; Image^.HeaderSize:=HeaderSize;
      if AHeader<>nil
      then Move(AHeader^, Header^, HeaderSize)
      else FillChar(Header^, HeaderSize, 0);
      if FormatId<>nil then Move(FormatId^[1], Header^, Length(FormatId^));
      if FSizeOfs>=0 then
      begin
         FileSizeH:=Image^.FileSize-4-FSizeInc;
         Move(FileSizeH, Header^[FSizeOfs], 2);
      end;
      ImageSizeH.X:=Image^.Size.X-SizesInc;
      ImageSizeH.Y:=Image^.Size.Y-SizesInc;
      Move(ImageSizeH.X, Header^[SizesOfs], 2);
      Move(ImageSizeH.Y, Header^[SizesOfs+SizesSize], 2);
   end;
   Image^.FormatId:=FormatId;
   ConvertImage:=True;
end;

procedure TGEditor.LoadImage;
var
   S               :TBufStream;
   FileSize        :LongInt;
   ImageSizeX,
   ImageSizeY,
   ImageSizeB      :LongInt;
   HeaderSize,
   i               :Integer;
   Header          :PByteArray;
   PrevImage       :PGEditorImage;
   Id              :String[31];
   NewTitle        :PString;

procedure LoadPCX(var Buffer;
                  Count                :LongInt;
                  Limit                :Word);
var
   LineBuf         :Array [0..255] of Byte;
   ReadCount       :Word;
   Extra           :Boolean;

procedure ReadByte;
begin
   S.Read(LineBuf, 1);
   Dec(Count);
end;

begin
   Inc(Limit, Ofs(Buffer));
   Extra:=False;

   repeat
      ReadCount:=SizeOf(LineBuf);
      if ReadCount>Count then ReadCount:=Count;
      S.Read(LineBuf, ReadCount);
      if S.Status<>stOk then Exit;
      Dec(Count, ReadCount);

      asm
         lea  si,LineBuf
         mov  cx,ReadCount
         les  di,Buffer
         mov  bx,Limit
         cld

@@NextByte:
         segss
         lodsb
         mov  dl,al
         and  dl,0C0h
         cmp  dl,0C0h
         jne  @@NoRepeat
         and  al,3Fh
         cbw
         dec  cx
         jnz  @@DataHere
         inc  cx
         push es
         pusha
         push bp
         call ReadByte
         popa
         pop  es
         lea  si,LineBuf
@@DataHere:
         xchg ax,cx
         push ax
         mov  ax,bx
         sub  ax,di
         cmp  cx,ax
         jbe  @@Fits
         mov  cx,ax
         xor  bx,bx
@@Fits:
         segss
         lodsb
         rep  stosb
         pop  cx
         cmp  di,bx
         ja   @@Stop
         loop @@NextByte
         jmp  @@Done
@@NoRepeat:
         cmp  di,bx
         jae  @@Stop
         stosb
         loop @@NextByte
         jmp  @@Done
@@Stop:
         not  Extra
@@Done:
         mov  word ptr Buffer,di
      end;

      if Extra then
      begin
         MessageBox(^M^C'Extra data in PCX, ignored.', nil, mfWarning+mfOKButton);
         Break;
      end;
   until Count=0;

   if Ofs(Buffer)<Limit then
      MessageBox(^M^C'Not enough data in PCX.', nil, mfWarning+mfOKButton);
end;

begin
   if Title=nil then Exit;
   S.Init(Title^, stOpenRead, FileBuffer);

   if (S.GetSize<MaxHeaderSize) and (S.GetSize>0)
   then HeaderSize:=S.GetSize else HeaderSize:=MaxHeaderSize;

   Header:=MemAlloc(HeaderSize);
   if Header=nil then
   begin
      Application^.OutOfMemory; S.Done; Exit;
   end;
   S.Read(Header^, HeaderSize);
   Move(Header^, Id[1], SizeOf(Id)-1);

   i:=1;
   while i<=FormatsCount do
   with Formats[i] do
   begin
      if FormatId<>nil then Id[0]:=FormatId^[0];

      if FSizeOfs<0 then FileSize:=S.GetSize else
      begin
         FileSize:=0; Move(Header^[FSizeOfs], FileSize, FSizeSize);
      end;
      Inc(FileSize, FSizeInc);

      ImageSizeX:=0; Move(Header^[SizesOfs], ImageSizeX, SizesSize);
      ImageSizeY:=0; Move(Header^[SizesOfs+SizesSize], ImageSizeY, SizesSize);
      Inc(ImageSizeX, SizesInc); Inc(ImageSizeY, SizesInc);
      ImageSizeB:=ImageSizeX*ImageSizeY;

      if ((FormatId=nil) or (Id=FormatId^)) and
         (FileSize>0) and (ImageSizeX>0) and (ImageSizeY>0) then
      if (Packing<>ipNone) or (FileSize=ImageSizeB) then Break else
      if Multiple and ((FileSize-FSizeInc) mod (ImageSizeB-FSizeInc) =0) then
      begin
         S.Seek(FileSize-ImageSizeB);
         S.Read(Header^, HeaderSize);

         if (PWord(@Header^[SizesOfs])^+SizesInc=Word(ImageSizeX)) and
            (PWord(@Header^[SizesOfs+SizesSize])^+SizesInc=Word(ImageSizeY)) then Break;
      end;

      Inc(i);
   end;
   FreeMem(Header, HeaderSize);

   if (S.Status=stOk) and (i>FormatsCount) then
   begin
      S.Done; MessageBox(^M^C'Unknown image format.', nil, mfError+mfOKButton);
      Exit;
   end;

   if (ImageSizeX>MaxImageSizeX) or (ImageSizeY>MaxImageSizeY) or
      (ImageSizeX*ImageSizeY+4>MaxImageSize) then
   begin
      S.Done; MessageBox(^M^C'Image is too large.', nil, mfError+mfOKButton);
      Exit;
   end;

   if S.Status=stOk then
   begin
      FreeImages;

      HeaderSize:=Formats[i].HeaderSize;
      S.Seek(0);

      PrevImage:=nil;
      repeat
         Header:=MemAlloc(HeaderSize);
         if Header=nil then
         begin
            S.Done; FreeImages;
            Application^.OutOfMemory; Exit;
         end;

         S.Read(Header^, HeaderSize);

         FileSize:=ImageSizeX*ImageSizeY;
         Image:=MemAlloc(FileSize + ImageHeaderSize);
         if Image=nil then
         begin
            FreeMem(Header, HeaderSize); S.Done; FreeImages;
            Application^.OutOfMemory; Exit;
         end;

         Image^.Clear;
         if PrevImage=nil then
         begin
            FirstImage:=Image; Image^.Prev:=Image; Image^.Next:=Image;
         end else
         begin
            PrevImage^.Next:=Image; Image^.Prev:=PrevImage;
            Image^.Next:=FirstImage; FirstImage^.Prev:=Image;
         end;
         PrevImage:=Image;

         Image^.FileSize:=FileSize+4;
         Image^.Size.X:=ImageSizeX; Image^.Size.Y:=ImageSizeY;
         Image^.Header:=Header; Image^.HeaderSize:=HeaderSize;
         case Formats[i].Packing of
            ipNone:
               S.Read(Image^.Data, FileSize);
            ipPCX:
               LoadPCX(Image^.Data, S.GetSize+Formats[i].FSizeInc, FileSize);
         end;
         Image^.FormatId:=i;

         if (S.Status=stOk) and Formats[i].Reverse then
         if not Image^.Effect(cmFlipY) then
         begin
            S.Done; FreeImages; Application^.OutOfMemory; Exit;
         end;
      until (S.GetPos=S.GetSize) or (not Formats[i].Multiple);
      Image:=FirstImage;
   end;

   with Formats[i] do
   if PaletteOfs<>-1 then
   begin
      if PaletteOfs>0 then S.Seek(PaletteOfs) else S.Seek(S.GetSize+PaletteOfs);
      LoadPalette(S, PaletteFmt);
   end;

   S.Done;
   if S.Status<>stOk then
   begin
      FreeImages; MessageBox(^M^C'Unable to load image.', nil, mfError+mfOKButton); Exit;
   end;

   if Formats[i].Packing<>ipNone then
   begin
      if not ConvertImage(1, nil) then
      begin
         FreeImages; Exit;
      end;

      NewTitle:=NewStr(Copy(Title^, 1, Pos('.', Title^))+SpriteExt);
      DisposeStr(Title); Title:=NewTitle;
   end;
end;

function  TGEditor.SaveImage;
var
   S               :TBufStream;
   Ok              :Boolean;
   Cur, Start      :PGEditorImage;
begin
   SaveImage:=False; Ok:=True;
   if FileName=nil then Exit;

   S.Init(FileName^, stCreate, FileBuffer);

   Start:=FirstImage; Cur:=Start;
   repeat
      S.Write(Cur^.Header^, Cur^.HeaderSize);

      if Formats[Cur^.FormatId].Reverse then
      begin
         Ok:=Cur^.Effect(cmFlipY);
         if not Ok then Application^.OutOfMemory;
      end;
      S.Write(Cur^.Data, Cur^.FileSize-4);

      Cur:=Cur^.Next;
   until Cur=Start;

   S.Done;

   if Formats[Image^.FormatId].Reverse and Ok then Image^.Effect(cmFlipY);

   if (S.Status<>stOk) or (not Ok)
   then MessageBox(^M^C'Unable to save image.', nil, mfError+mfOKButton) else
   begin
      repeat Cur^.Modified:=False; Cur:=Cur^.Next until Cur=Start;
      Field^.ListModified:=False;

      SaveImage:=True;
   end;
end;

{ TColorView }

procedure TColorView.DrawColor;
var
   P               :TGPoint;
   S               :String[7];
   TextColor       :Integer;
begin
   HideMouseLocal(Bounds);
   DrawBar(Bounds, Color);
   Str(Color, S);
   P.X:=Bounds.CenterX; P.Y:=Bounds.CenterY + 1;
   if PaletteRGB[Color, 0]+PaletteRGB[Color, 1]+PaletteRGB[Color, 2] > 3*64 div 2
   then TextColor:=GetColor(cpDark) else TextColor:=GetColor(cpLight);
   DrawText(P, S, jsCenter, nil, TextColor);
   ShowMouseRect;
end;

{ TPaletteBox }

constructor TPaletteBox.Init;
begin
   Right.A.X:=Right.B.X-PaletteGrid.X*PaletteCount.X-1;
   Inherited Init(Right);
   EventMask:=EventMask or evMouseMove;
   CLeft:=GetColor(cpMouseLeft); CRight:=GetColor(cpMouseRight);
   CCurrent:=-1;
end;

function  TPaletteBox.GetPalette;
begin
   GetPalette:=CPaletteBox;
end;

procedure TPaletteBox.Draw;
var
   Pos             :TGPoint;
   R, C            :TGRect;
   Color, ColorU   :Byte;
   Colors          :Record
      Size         :TGPoint;
      Image        :Array [0..255] of Byte;
   end;
begin
   Color:=0;
   Colors.Size:=PaletteCount;
   for Pos.X:=0 to PaletteCount.X-1 do
   for Pos.Y:=0 to PaletteCount.Y-1 do
   begin
      Colors.Image[Pos.Y * PaletteCount.X + Pos.X]:=Color; Inc(Color);
   end;

   GetExtent(R);
   DrawRect(R, GetColor(cpFrame));
   R.Grow(-1, -1);
   DrawImage(R.A, PaletteGrid, @Colors, NullColor);

   if PaletteBox^.UsageShown then
   begin
      Color:=0;
      for Pos.X:=0 to PaletteCount.X-1 do
      for Pos.Y:=0 to PaletteCount.Y-1 do
      begin
         if Color in Used then
         begin
            if PaletteRGB[Color, 0]+PaletteRGB[Color, 1]+PaletteRGB[Color, 2] > 3*64 div 2
            then ColorU:=GetColor(cpDark) else ColorU:=GetColor(cpLight);

            C.A.X:=R.A.X+Pos.X*PaletteGrid.X; C.A.Y:=R.A.Y+Pos.Y*PaletteGrid.Y;
            C.SetSize(PaletteGrid);
            C.Grow(-1, -1);
            DrawRect(C, ColorU);
         end;
         Inc(Color);
      end;
   end;

   R.A.Y:=PaletteCount.Y * PaletteGrid.Y + 1;
   DrawBar(R, GetColor(cpMain));

   DrawCurrent(CCurrent);
end;

procedure TPaletteBox.DrawCurrent;
var
   R               :TGRect;
begin
   GetExtent(R); R.Grow(-1, -1); R.A.Y:=R.B.Y - CharHeight - 2;
   if Color<0 then DrawBar(R, GetColor(cpMain)) else
   begin
      DrawRect(R, GetColor(cpDark)); R.Grow(-1, -1);
      DrawColor(R, Color);
   end;
   CCurrent:=Color;
end;

procedure TPaletteBox.Update;
var
   Extent          :TGRect;
begin
   if CCurrent<0 then Exit;
   GetExtent(Extent);
   MakeLocal(Mouse, Mouse);
   if not Extent.Contains(Mouse) then DrawCurrent(-1);
end;

procedure TPaletteBox.HandleEvent;
var
   MouseColor      :Integer;

function  CalcColor                    :Integer;
var
   Pos             :TGPoint;
begin
   MakeLocal(Event.Where, Event.Where); Event.Where.Move(-1, -1);
   Pos.X:=Event.Where.X div PaletteGrid.X; Pos.Y:=Event.Where.Y div PaletteGrid.Y;
   if (Pos.X<0) or (Pos.X>=PaletteCount.X) or
      (Pos.Y<0) or (Pos.Y>=PaletteCount.Y) then CalcColor:=-1
   else CalcColor:=Pos.X*PaletteCount.Y+Pos.Y;
   ClearEvent(Event);
end;

procedure SetColor(var Color           :Integer);
var
   Last            :Integer;
begin
   Last:=Color; Color:=CalcColor;
   if Color<0 then Color:=Last else
   if Color<>Last then InfoBar^.DrawView;
end;

begin
   Inherited HandleEvent(Event);
   if Event.What and evMouseDown<>0 then
   if Event.Buttons and mbRightButton=0
   then SetColor(CLeft) else SetColor(CRight) else
   if Event.What and evMouseMove<>0 then
   begin
      MouseColor:=CalcColor;
      if MouseColor<>CCurrent then DrawCurrent(MouseColor);
   end;
end;

{ TInfoBar }

function  TInfoBar.GetPalette;
begin
   GetPalette:=CInfoBar;
end;

procedure TInfoBar.Draw;
var
   R               :TGRect;
   P               :TGPoint;
   Op              :Char;
   Vx, Vy          :String[5];
begin
   GetExtent(R);

   DrawButton(R, dbFillIn+dbSmall, False, GetColor(cpMain));
   DrawPos;

   case EditMode of
      cmSetColor:  Op:='S';
      cmGetColor:  Op:='G';
      cmFillArea:  Op:='F';
      cmReplace:   Op:='r';
      cmRectangle: Op:='R';
      cmBar:       Op:='B';
      cmLine:      Op:='L';
      cmCopy:      Op:='C';
      cmPaste:     Op:='P';
   else
      Op:=' ';
   end;
   P.X:=19*CharWidth; P.Y:=R.CenterY + 1;
   DrawText(P, Op, jsCenterY, nil, GetColor(cpText));

   R.Grow(-1, -1); R.A.X:=R.B.X - 3*CharWidth - 3;
   DrawColor(R, PaletteBox^.CRight);
   R.Move(-R.SizeX-1, 0);
   DrawColor(R, PaletteBox^.CLeft);

   Str(DriverPtr^.SizeX, Vx);
   Str(DriverPtr^.SizeY, Vy);
   P.X:=R.A.X-CharWidth;
   DrawText(P, Vx+'x'+Vy, jsRight+jsCenterY, nil, GetColor(cpText));
end;

procedure TInfoBar.DrawPos;
var
   R               :TGRect;
   S               :String[23];
   P               :Record
      Modified,
      PosX, PosY,
      Num, Cnt     :LongInt;
   End;
begin
   if not Exposed then Exit;
   GetExtent(R);
   R.Grow(-1, -1);
   R.A.X:=8; R.B.X:=R.A.X + 17*CharWidth -1;

   WaitRetrace;

   HideMouseLocal(R);
   DrawBar(R, GetColor(cpMain));
   if CorrectData then
   begin
      if Modified then Char(P.Modified):=#15 else Char(P.Modified):=' ';
      P.PosX:=Pos.X; P.PosY:=Pos.Y;
      P.Num:=ImageNum; P.Cnt:=ImageCnt;

      FormatStr(S, '%c%4d:%-3d%4d/%-3d', P);

      R.A.Y:=R.CenterY + 1;
      DrawText(R.A, S, jsCenterY, nil, GetColor(cpText));

      if Color<>PaletteBox^.CCurrent then PaletteBox^.DrawCurrent(Color);
   end;
   ShowMouseRect;
end;

begin
   MaxBufMem:=MaxHeapSize;
end.
